#include <LiquidCrystal.h>
#include <LCDKeypad.h>
#include <OneWire.h>
#include <DallasTemperature.h>
#include <MenuBackend.h>
#include <EEPROMex.h>


//eeprom addresses
    int EEBCKL = EEPROM.getAddress(sizeof(boolean));
    int EEBUZZ = EEPROM.getAddress(sizeof(boolean));
    int EEREFIL = EEPROM.getAddress(sizeof(unsigned char));
    int EETHOT = EEPROM.getAddress(sizeof(float));
    int EETCOLD = EEPROM.getAddress(sizeof(float));
    int EETHIST = EEPROM.getAddress(sizeof(float));
    int EEWINTER = EEPROM.getAddress(sizeof(boolean));
    int EEAUTOFILL = EEPROM.getAddress(sizeof(boolean));
    int EETBOILERERROR = EEPROM.getAddress(sizeof(float));
    int EEFLAG = 0xFF;

/*
#define EEBCKL 0
#define EEBUZZ 1
#define EEREFIL 2
#define EETHOT 3
#define EETCOLD 4
#define EETHIST 5
#define EEFLAG  0xFF
*/

//pin definitions
#define TCOL A1 //DS18B20 inside collector
#define TBOIL A5 //DS18B20 inside boiler
#define L1 A2 //water level pins
#define L2 A3
#define L3 A4
#define BUZZER 13 
#define VLV 2 //valve control pin
#define PMP 3 //circular pump control pin
#define HTR 11 // heater control pin
//sensor ground
#define SGND 13
//parameter limits
#define HISTMIN 0.0
#define HISTMAX 10.0
#define THOTMIN 50.0
#define THOTMAX 100.0
#define TCOLDMIN 0.0
#define TCOLDMAX 10.0
#define LCDBCKL 10.0
//constants
#define BKLDELTA 10
//menu update
#define TIMER1CNT 2
//sensors update
#define TIMER1CNT1 10
//peripherals
OneWire ds(TCOL); //attach sensor inside colector
OneWire ds1(TBOIL); //attach sensor inside colector
DallasTemperature sensors(&ds);
DallasTemperature sensors1(&ds1);
LCDKeypad lcd; //attach LCD keypad shield
//variables
boolean heater = 0; //collector water heater 0 - off
boolean pump = 0; //water pump 0 - off
boolean valve = 0; // water filling valve 0- off
//collector water temperature
float tcollector;
//boiler water temperature
float tboiler;
//water level value 1; 2; 3;  0 error
unsigned char wlevel = 1; //water level 1 - lowest
//backlight flag
boolean bckl = 1;
//buzzer flag
boolean buzz = 1;
//threshold values in degrees
float thot = 70.0;
float tcold = 5.0;
float thist = 3.0;
//boiler temperature measure error
float tboilerror = 8;
//refil activation level
unsigned char refillvl = 2;
//winter flag
boolean winterflag = 0;
//autofil flag
boolean autofillflag = 1;
//test pump
boolean testpumpflag = 0;

//button pressed value
unsigned int lastButton = KEYPAD_NONE;
//timer controlled variables
boolean timer1out = 0;
unsigned char timer1out1 = 0;
//this controls the menu backend and the event generation
MenuBackend menu = MenuBackend(menuUseEvent, menuChangeEvent); 
	//beneath is list of menu items needed to build the menu
	MenuItem tempcold = MenuItem("TempCold");
        MenuItem temphot = MenuItem("TempHot");
        MenuItem temphist = MenuItem("TempHist");
	MenuItem refil = MenuItem("Refil LVL");
	MenuItem bcklt = MenuItem("Backlight");
        MenuItem buzzer = MenuItem("Buzzer");
        MenuItem winter = MenuItem("Winter");
        MenuItem autofill = MenuItem("Auto fill");
        MenuItem testpump = MenuItem("Test pump");
        MenuItem tboilererror = MenuItem("Boiler T Error");       

//this function builds the menu and connects the correct items together
void menuSetup()
{
	//add the tempcold menu to the menu root
	menu.getRoot().add(tempcold); 
		//we want looping both up and down
                tempcold.addBefore(tempcold);		
                //tempcold.addBefore(buzzer);
		tempcold.addAfter(temphot);
                temphot.addAfter(temphist);
                temphist.addAfter(refil);
                refil.addAfter(bcklt);                        
		bcklt.addAfter(buzzer);
                buzzer.addAfter(winter);
                winter.addAfter(testpump);
                testpump.addAfter(tboilererror);
                tboilererror.addAfter(autofill);
                autofill.addAfter(tempcold);

}
/*
	This is an important function
	Here all use events are handled
	This is where you define a behaviour for a menu item
*/
void menuUseEvent(MenuUseEvent used)
{

	if (used.item == tempcold) //comparison against a known item
	{
		float ttcold = tcold;
                lcd.clear();
                lcd.print("<<TempCold>>");
                do{
                  buttonRead();
                  if (lastButton == KEYPAD_LEFT){
                    if (ttcold - 1.0 > TCOLDMIN)
                      ttcold -=  1.0;
                    else
                      ttcold = TCOLDMIN;
                  }
                  if (lastButton == KEYPAD_RIGHT){
                    if (ttcold + 1.0 < TCOLDMAX)
                      ttcold +=  1.0;
                    else
                      ttcold = TCOLDMAX;
                  } 
                  lcd.setCursor(0,1);  
                  lcd.print(ttcold,1);       
                  lcd.print("C  ");                
                }while (lastButton != KEYPAD_SELECT); 
                tcold = ttcold;
	}
	if (used.item == temphot) //comparison against a known item
	{
		lcd.clear();
                lcd.print("<<TempHot>>");
                float tthot = thot;
                do{
                  buttonRead();
                  if (lastButton == KEYPAD_LEFT){
                    if (tthot - 1.0 > THOTMIN)
                      tthot -=  1.0;
                    else
                      tthot = THOTMIN;
                  }
                  if (lastButton == KEYPAD_RIGHT){
                    if (tthot + 1.0 < THOTMAX)
                      tthot +=  1.0;
                    else
                      tthot = THOTMAX;
                  } 
                  lcd.setCursor(0,1);  
                  lcd.print(tthot,1);       
                  lcd.print("C  ");                
                }while (lastButton != KEYPAD_SELECT); 
                thot = tthot;                

	}
	if (used.item == temphist) //comparison against a known item
	{
		lcd.clear();
                lcd.print("<<TempHist>>");
                float tthist = thist;
                do{
                  buttonRead();
                  if (lastButton == KEYPAD_LEFT){
                    if (tthist - 1.0 > HISTMIN)
                      tthist -=  1.0;
                    else
                      tthist = HISTMIN;
                  }
                  if (lastButton == KEYPAD_RIGHT){
                    if (tthist + 1.0 < HISTMAX)
                      tthist +=  1.0;
                    else
                      tthist = HISTMAX;
                  } 
                  lcd.setCursor(0,1);  
                  lcd.print(tthist,1);       
                  lcd.print("C  ");                
                }while (lastButton != KEYPAD_SELECT);                 
                thist = tthist;
                

	}
	if (used.item == refil) //comparison against a known item
	{
                unsigned char trefillvl = refillvl;		
                lcd.clear();
                lcd.print("<<Refil Level>>");

                do{
                  buttonRead();
                  if (lastButton == KEYPAD_LEFT){
                    if (trefillvl > 1)
                      trefillvl--;
                    else
                      trefillvl = 1;
                  }
                  if (lastButton == KEYPAD_RIGHT){
                    if (trefillvl < 3)
                      trefillvl++;
                    else
                      trefillvl = 3;
                  } 
                  lcd.setCursor(0,1);   
                  if (trefillvl ==1)
                   lcd.print("      *--      ");  
                  else if (trefillvl ==2)
                   lcd.print("      **-      ");                        
                  else
                   lcd.print("      ***      ");
                }while (lastButton != KEYPAD_SELECT); 
                refillvl = trefillvl;
	}
	if (used.item == bcklt) //comparison against a known item
	{
		boolean tbckl = bckl;
                lcd.clear();
                lcd.print("<<Backlight>>");
                do{
                  buttonRead();
                  if (lastButton == KEYPAD_LEFT){
                    tbckl = 0;
                  }
                  if (lastButton == KEYPAD_RIGHT){
                    tbckl = 1;
                  } 
                  lcd.setCursor(0,1);        
                  if (tbckl ==1)
                   lcd.print("      ON       ");  
                  else
                   lcd.print("      OFF      ");           
                 }while (lastButton != KEYPAD_SELECT);
                 bckl = tbckl;
                 digitalWrite(LCDBCKL, bckl);
             
	}
	if (used.item == buzzer) //comparison against a known item
	{
		boolean tbuzz = buzz;
                lcd.clear();
                lcd.print("<<Buzzer>>");
                do{
                  buttonRead();
                  if (lastButton == KEYPAD_LEFT){
                    tbuzz = 0;
                  }
                  if (lastButton == KEYPAD_RIGHT){
                    tbuzz = 1;
                  } 
                  lcd.setCursor(0,1);        
                  if (tbuzz ==1)
                   lcd.print("      ON       ");  
                  else
                   lcd.print("      OFF      ");           
                 }while (lastButton != KEYPAD_SELECT);                
                buzz = tbuzz;

	}
	if (used.item == winter) //comparison against a known item
	{
		boolean twinter = winterflag;
                lcd.clear();
                lcd.print("<<Winter>>");
                do{
                  buttonRead();
                  if (lastButton == KEYPAD_LEFT){
                    twinter = 0;
                  }
                  if (lastButton == KEYPAD_RIGHT){
                    twinter = 1;
                  } 
                  lcd.setCursor(0,1);        
                  if (twinter ==1)
                   lcd.print("      ON       ");  
                  else
                   lcd.print("      OFF      ");           
                 }while (lastButton != KEYPAD_SELECT);                
                winterflag = twinter;

	}
	if (used.item == autofill) //comparison against a known item
	{
		boolean tautofill = autofillflag;
                lcd.clear();
                lcd.print("<<Auto fill>>");
                do{
                  buttonRead();
                  if (lastButton == KEYPAD_LEFT){
                    tautofill = 0;
                  }
                  if (lastButton == KEYPAD_RIGHT){
                    tautofill = 1;
                  } 
                  lcd.setCursor(0,1);        
                  if (tautofill ==1)
                   lcd.print("      ON       ");  
                  else
                   lcd.print("      OFF      ");           
                 }while (lastButton != KEYPAD_SELECT);                
                autofillflag = tautofill;

	}
	if (used.item == testpump) //comparison against a known item
	{
		//boolean tautofill = autofillflag;
                lcd.clear();
                lcd.print("<<Test pump>>");
                do{
                  buttonRead();
                  if (lastButton == KEYPAD_LEFT){
                    testpumpflag = 0;
                  }
                  if (lastButton == KEYPAD_RIGHT){
                    testpumpflag = 1;
                  } 
                  lcd.setCursor(0,1);        
                  if (testpumpflag ==1)
                   lcd.print("      ON       ");  
                  else
                   lcd.print("      OFF      ");           
                 }while (lastButton != KEYPAD_SELECT);                
                //autofillflag = tautofill;

	}
	if (used.item == tboilererror) //comparison against a known item
	{
		float ttboilerror = tboilerror;
                lcd.clear();
                lcd.print("<<dT Boiler>>");
                do{
                  buttonRead();
                  if (lastButton == KEYPAD_LEFT){
                    if (ttboilerror - 1.0 > 0)
                      ttboilerror -=  1.0;
                    else
                      ttboilerror = 0;
                  }
                  if (lastButton == KEYPAD_RIGHT){
                    if (ttboilerror + 1.0 < 30)
                      ttboilerror +=  1.0;
                    else
                      ttboilerror = 30;
                  } 
                  lcd.setCursor(0,1);  
                  lcd.print(ttboilerror,1);       
                  lcd.print("C  ");                
                }while (lastButton != KEYPAD_SELECT); 
                tboilerror = ttboilerror;
	}


lcd.setCursor(14,1); 
lcd.print("OK"); 
eepromUpdate();
timer1out = 0;
}

void menuChangeEvent(MenuChangeEvent changed)
{
  lcd.clear();
  lcd.print(changed.to.getName());
  lcd.setCursor(0,1);  
  if (changed.to.getName()=="Backlight")
    if (bckl ==1)
     lcd.print("      ON       "); 
    else
     lcd.print("      OFF      ");    
  if (changed.to.getName()=="Buzzer")
    if (buzz ==1)
     lcd.print("      ON       "); 
    else
     lcd.print("      OFF      ");   
  if (changed.to.getName()=="Refil LVL")
    if (refillvl ==1)
      lcd.print("      *--      ");
    else if (refillvl ==2)
     lcd.print("      **-      "); 
    else
     lcd.print("      ***      ");    
  if (changed.to.getName()=="TempCold") 
    {
      lcd.print(tcold,1);       
      lcd.print("C ");           
    }
  if (changed.to.getName()=="TempHot") 
    {
      lcd.print(thot,1);       
      lcd.print("C ");           
    }
  if (changed.to.getName()=="TempHist") 
    {
      lcd.print(thist,1);       
      lcd.print("C ");           
    }
  if (changed.to.getName()=="Winter")
    if (winterflag ==1)
     lcd.print("      ON       "); 
    else
     lcd.print("      OFF      ");   
  if (changed.to.getName()=="Auto fill")
    if (autofillflag ==1)
     lcd.print("      ON       "); 
    else
     lcd.print("      OFF      ");   
  if (changed.to.getName()=="Test pump")
    if (testpumpflag ==1)
     lcd.print("      ON       "); 
    else
     lcd.print("      OFF      ");   
  if (changed.to.getName()=="Boiler T Error") 
    {
      lcd.print(tboilerror,1);       
      lcd.print("C ");           
    }
} 
void eepromRestore()
{
  
  if (EEPROM.read(EEFLAG)!=0x55){//if first time eeprom access, write initial values
    EEPROM.write(EEBCKL, bckl);
    EEPROM.write(EEBUZZ, buzz);
    EEPROM.write(EEREFIL, refillvl);
    EEPROM.writeFloat(EETHOT, thot);
    EEPROM.writeFloat(EETCOLD, tcold);
    EEPROM.writeFloat(EETHIST, thist);
    EEPROM.write(EEWINTER, winterflag);
    EEPROM.write(EEAUTOFILL, autofillflag);
    EEPROM.write(EETBOILERERROR, tboilerror);
    EEPROM.write(EEFLAG, 0x55);
  }
  bckl = EEPROM.read(EEBCKL);
  buzz = EEPROM.read(EEBUZZ);
  refillvl = EEPROM.read(EEREFIL);
  thot = EEPROM.readFloat(EETHOT);
  tcold = EEPROM.readFloat(EETCOLD);
  thist = EEPROM.readFloat(EETHIST);
  winterflag = EEPROM.read(EEWINTER);
  autofillflag = EEPROM.read(EEAUTOFILL);  
  tboilerror = EEPROM.read(EETBOILERERROR);  
}
void eepromUpdate(){
  if (EEPROM.read(EEBCKL) != bckl)
    EEPROM.write(EEBCKL, bckl);
  if (EEPROM.read(EEBUZZ) != buzz)
    EEPROM.write(EEBUZZ, buzz);
  if (EEPROM.read(EEREFIL) != refillvl)
    EEPROM.write(EEREFIL, refillvl);
  if(EEPROM.readFloat(EETHOT != thot))
    EEPROM.writeFloat(EETHOT, thot);
  if(EEPROM.readFloat(EETCOLD != tcold))
    EEPROM.writeFloat(EETCOLD, tcold);
  if(EEPROM.readFloat(EETHIST != thist))
    EEPROM.writeFloat(EETHIST, thist);  
  if (EEPROM.read(EEWINTER) != winterflag)
    EEPROM.write(EEWINTER, winterflag);
  if (EEPROM.read(EEAUTOFILL) != autofillflag)
    EEPROM.write(EEAUTOFILL, autofillflag);
  if (EEPROM.read(EETBOILERERROR) != tboilerror)
    EEPROM.write(EETBOILERERROR, tboilerror);  
  
  //update backup values to eeprom
}
void setup(){
  pinMode(L1, INPUT);
  pinMode(L2, INPUT);
  pinMode(L3, INPUT);
  pinMode(LCDBCKL, OUTPUT);
  pinMode(VLV, OUTPUT);
  pinMode(PMP, OUTPUT);
  pinMode(HTR, OUTPUT);
  pinMode(SGND, OUTPUT);
  digitalWrite(SGND, HIGH);
  eepromRestore();
  digitalWrite(LCDBCKL, bckl);
  lcd.begin(16,2);
  lcd.clear();
  lcd.print("Solar Controller");  
  lcd.setCursor(0,1);
  lcd.print("      v1.0    ");
  delay(500);
  sensors.begin();//IC default 9 bit
  sensors1.begin();//IC default 9 bit
  menuSetup(); 

  cli();
  //set timer1 interrupt every ~4 seconds 
  TCCR1A = 0;// set entire TCCR1A register to 0
  TCCR1B = 0;// same for TCCR1B
  TCNT1  = 0;//initialize counter value to 0
  // set compare match register for 0.25hz increments
  OCR1A = 0xFFFF;// 
  // turn on CTC mode
  TCCR1B |= (1 << WGM12);
  // Set CS12 and CS10 bits for 1024 prescaler
  TCCR1B |= (1 << CS12) | (1 << CS10);  
  // enable timer compare interrupt
  TIMSK1 |= (1 << OCIE1A);
  sei();
  
}
void loop(){
  buttonRead();
  switch (lastButton) {
	case KEYPAD_UP: 
          menu.moveUp();
          timer1out = TIMER1CNT;
          timer1out1 = 0;
          break;
	case KEYPAD_DOWN:
          menu.moveDown();
          timer1out = TIMER1CNT;
          timer1out1 = 0;
          break;
	case KEYPAD_RIGHT:
          menu.moveRight();
          timer1out = TIMER1CNT;
          timer1out1 = 0;
          break;
	case KEYPAD_LEFT:
          menu.moveLeft();
          timer1out = TIMER1CNT;
          timer1out1 = 0;
          break;
	case KEYPAD_SELECT:
          menu.use();
          timer1out = TIMER1CNT;
          timer1out1 = 0;
          break;
        //case KEYPAD_NONE: menu.use(); break;
		} 
  lastButton = KEYPAD_NONE;

  if (timer1out1 == 0)
  {
    digitalWrite(SGND, LOW);
    sensors.begin();//IC default 9 bit
    sensors1.begin();//IC default 9 bit
    readTemperatures();
    getWaterLevel();
    digitalWrite(SGND, HIGH);
    timer1out1 = TIMER1CNT1;
  }
  if (timer1out == 0)
  {

    updateStatus();
    mainView();
    timer1out = TIMER1CNT;
  }
}

ISR(TIMER1_COMPA_vect){//timer1 interrupt clears flag
if (timer1out >=1 )
  timer1out--;
  else timer1out = 0;
if (timer1out1 >=1 )
  timer1out1--;
  else timer1out1 = 0;
}

void mainView(){
  
  //lcd.clear();
  lcd.setCursor(0,0);
  lcd.print("S:");
  lcd.print(tcollector,1);
  lcd.print("C; ");
  lcd.print("B:");
  lcd.print(tboiler,1);
  lcd.print("C; ");
  lcd.setCursor(0,1);
  lcd.print("L:");
    if (wlevel == 0)
      lcd.print("---");
      else if (wlevel==1)
        lcd.print("*--");
        else if (wlevel ==2)
          lcd.print("**-");
          else
              lcd.print("***");    
 if (winterflag ==1)
 {
   lcd.print("Winter");
 }
 else{
  lcd.print("P:");
    if (pump)
      lcd.print("1");
      else 
        lcd.print("0");
  lcd.print(" V:");
      if (valve)
        lcd.print("1");
      else 
        lcd.print("0");
 }
  lcd.print(" H:");
      if (heater)
        lcd.print("1");
      else 
        lcd.print("0");  
}
/*

*/
void buttonRead(){
  delay(50);
  lastButton = lcd.button();
  //wait for button release
 while(lcd.button()!=KEYPAD_NONE)
 {
 }  
}
/*
read collector and boiler temperatures 
*/
void readTemperatures()
{
 sensors.requestTemperatures(); // Send the command to get temperatures
 sensors1.requestTemperatures(); // Send the command to get temperatures
 delay(500);
 tcollector = sensors.getTempCByIndex(0);
 tboiler =  sensors1.getTempCByIndex(0) + tboilerror;
}
void getWaterLevel(){
  //read water level
  if (digitalRead(L3))
    {
      wlevel = 3;
    } else if (digitalRead(L2))
      {
          wlevel = 2;
      } else if (digitalRead(L1))
        {
          wlevel = 1;
        } else {
          //empty
          wlevel = 0;
        }
    
}
void turnValveOn(){
  //turn valve on
  digitalWrite(VLV, 1);
  valve = 1;
}
void turnValveOff(){
  //turn valve off
  digitalWrite(VLV, 0);
  valve = 0;
}
void turnPumpOn(){
  //turn pump on
  digitalWrite(PMP, 1);
  pump = 1;
}
void turnPumpOff(){
  //turn pump off
  digitalWrite(PMP, 0);
  pump = 0;
}
void turnHeaterOn(){
  //turn heater on
  digitalWrite(HTR, 1);
  heater = 1;
}
void turnHeaterOff(){
  //turn heater off
  digitalWrite(HTR, 0);
  heater = 0;
}
void updateStatus()
{

if (winterflag ==0)
{
  if (autofillflag ==1)
    {
        //open valve is level is lower than refil level set
      if (wlevel < refillvl)
      {
        turnValveOn();
        timer1out1 = 0;//refresh sensor readings
      } 
      else
        turnValveOff();//turn off when filled up to refil level
    }


if(wlevel >=1)
{
  if(testpumpflag == 1)
  {
    turnPumpOn();
  } else {
     //pump control
     if ((tcollector > (tboiler + thist))||(tcollector >= thot)||(tcollector <= tcold))
        {
          turnPumpOn();
     }else if ((tcollector <= tboiler)&&(tcollector < thot)||(tcollector > tcold))
        {
          turnPumpOff();  
        }
  }
}



}else {
  if (testpumpflag == 0)
  {
    turnValveOff();
    turnPumpOff(); 
  }
 
}

  //heater control
//  turn heater on if colector temperature is lower than tcold and there is water at least 1 level
  if ((tcollector < tcold)&&(wlevel>=1))
    turnHeaterOn();
  else
    turnHeaterOff();
}
